
from __future__ import print_function

import six
import os
import sys
import uuid
import time
import contextlib
import json
import getpass
import subprocess
import platform
import traceback

from kabaret.app._actor import Actor, Cmd, Cmds



#------ CMDS

class JobsCmds(Cmds):
    '''
    The Jobs actor manages a list of jobs and associated pool queues.
    blah blah...
    blah blah...
    blah blah documentation is awesome...
    blah blah...
    blah blah...
    '''

@JobsCmds.cmd
class Create_Flow_Job(Cmd):
    '''
    Create a Job that will execute an kabaret.flow.Object
    (the job<->object interface is still TBD)
    '''
    def _decode(
        self, 
        oid, pool, priority=100, paused=True,
        label=None, creator=None, owner=None,
        show_console=None,
        interpreter=None,
        worker_filename=None,
    ):
        self._oid = oid
        self._pool = pool
        self._priority = priority
        self._paused = paused
        self._label = label
        self._creator = creator
        self._owner = owner
        self._show_console = show_console
        self._interpreter = interpreter
        self._worker_filename = worker_filename

    def _execute(self):
        kwargs = {'oid':self._oid}
        if not self._show_console:
            kwargs['show_console'] = self._show_console
        if self._interpreter is not None:
            kwargs['interpreter'] = self._interpreter
        if self._worker_filename is not None:
            kwargs['worker_filename'] = self._worker_filename
        
        params = ((),kwargs)

        job = self.actor().create_job(
            'flow', self._label, self._creator, self._owner, params, self._pool, self._priority, self._paused,
        )
        return job.jid()

@JobsCmds.cmd
class Create_Job(Cmd):
    '''
    Create a Job in the given pool, with the given priority, with the given
    paused state, of the given job type and with the optionally given args and kwargs.
    '''
    def _decode(self, pool, priority, paused, job_type, *job_args, **job_kwargs):
        self._oid = oid
        self._pool = pool
        self._priority = priority
        self._paused = paused
        self._args = job_args
        self._kwargs = job_kwargs

    def _execute(self):
        job = self.actor().create_job(
            'flow', (self._args, self._kwargs), self._pool, self._priority, self._paused
        )
        return job.jid()

@JobsCmds.cmd
class Stop_Node(Cmd):
    '''
    Create a Job that will stop the Node picking it.
    If node_id is a pool name, the first Node in that poll will be stoped.
    If node_id is the exact identifier of an existing Node, this Node will
    pick the job and stop itself.
    '''
    def _decode(self, node_id, priority=100):
        self._node_id = node_id
        self._priority = priority

    def _execute(self):
        params = ((),dict(cmd='stop'))
        job = self.actor().create_job(
            'NODE_CMD', params, self._node_id, self._priority, False
        )  
        return job.jid()

@JobsCmds.cmd
class Kill_Node_Job(Cmd):
    '''
    Create a Job that will kill the current job of the Node picking it.
    If node_id is a pool name, the first Node in that poll will kill its job.
    If node_id is the exact identifier of an existing Node, this Node will
    pick the job and kill its job.

    If the Node picking the job has no current running job, the job will
    be ignored but still marked as Done.
    '''

    def _decode(self, node_id, priority=100):
        self._node_id = node_id
        self._priority = priority

    def _execute(self):
        params = ((),dict(cmd='kill_job'))
        job = self.actor().create_job(
            'NODE_CMD', params, self._node_id, self._priority, False
        )
        return job.jid()


@JobsCmds.cmd
class List_Jobs(Cmd):
    '''
    Returns a list of dicts describing all the known jobs.
    The dict keys are strings:
        jid         : uniq id of the job
        created_on  : timestamp float
        job_tye     : job type string
        job_label     : job label string
        jog_params  : (args_list, kwargs_dict) 
        pool        : pool name string
        paused      : pauded state bool
        priority    : priority int
        node        : node id of the Node executing or executed the job
        in_progress : bool
        done        : bool

    '''
    def _decode(self):
        pass

    def _execute(self):
        jobs = self.actor().get_all_jobs()
        return [j.to_dict() for j in jobs ]

@JobsCmds.cmd
class Get_Job_Info(Cmd):
    '''
    Returns a dict describing the job with the given job id.
    See list_jobs cmd for dict key list.
    '''
    def _decode(self, jid):
        self._jid = jid

    def _execute(self):
        try:
            job = self.actor().get_job(self._jid)
        except ValueError:
            return Job.UnknownJobDict(self._jid)
        else:
            return job.to_dict()

@JobsCmds.cmd
class Set_Job_Paused(Cmd):
    '''
    Changes the paused status of the job with the given id.
    '''
    def _decode(self, jid, paused=True):
        self._jid = jid
        self._paused = paused

    def _execute(self):
        job = self.actor().get_job(self._jid)
        job.set_paused(self._paused)

@JobsCmds.cmd
class Set_Job_Error(Cmd):
    '''
    Sets the job error status and message
    '''

    def _decode(self, jid, error_message):
        self._jid = jid
        self._error_message = error_message

    def _execute(self):
        try:
            job = self.actor().get_job(self._jid)
        except ValueError:
            return
        else:
            job.set_error(self._error_message)

@JobsCmds.cmd
class Set_Job_Done(Cmd):
    '''
    Mark the job as done.
    '''

    def _decode(self, jid):
        self._jid = jid

    def _execute(self):
        try:
            job = self.actor().get_job(self._jid)
        except ValueError:
            return
        else:
            job.set_done()

@JobsCmds.cmd
class Set_Job_In_Progress(Cmd):
    '''
    Mark the job as in progress.
    '''

    def _decode(self, jid):
        self._jid = jid

    def _execute(self):
        try:
            job = self.actor().get_job(self._jid)
        except ValueError:
            return
        else:
            job.set_in_progress()


@JobsCmds.cmd
class Delete_Job(Cmd):
    '''
    Deletes the job with the given id.
    If the job is in progress, the forced argument must be True
    for the job to be deleted.
    '''
    def _decode(self, jid, forced=False):
        self._jid = jid
        self._forced = forced

    def _execute(self):
        return self.actor().delete_job(self._jid, self._forced)


class Job(object):
    '''
    Job fields are stored in the <CLUSTER_NAME>:Jobs:jobs:<jid> hash.
    The list of jobs is stored in the <CLUSTER_NAME>:Jobs:job_ids set.

    If not paused at creation time, its jid is added to the <CLUSTER_NAME>:Jobs:pool_q:<pool_name> sorted set
    
    Unpausing a job adds it to the pool sorted set with its priority as score.
    Pausing a job removes it from the pool sorted set.

    The workers takes jid from their affected pool sorted set and alter the job's hash with infos.
    '''

    @classmethod
    def create(cls, actor, job_type, job_label, job_creator, job_owner, job_params, pool, priority, paused):
        if job_creator is None:
            job_creator = actor.get_current_login()
        if job_owner is None:
            job_owner = actor.get_current_login()

        # Create an entry in the job_ids set
        created = 0
        max_tries = 10
        tried = 0
        while not created:
            tried += 1
            jid = str(uuid.uuid4())
            created = actor._db.sadd(actor.jobids_set_name, jid)
            if tried > max_tries:
                raise Exception('Could not create a new job id !!!!')

        # Store the job in the jobs hash
        field_created = actor._db.hsetnx(actor.job_hashes_namespace+'/'+jid, 'created_on', time.time())
        if not field_created:
            raise Exception(
                'Error: looks like this job is already defined ({})!!! Aborting job definition.'.format(
                    actor.job_hashes_namespace+'/'+jid
                )
            )

        job = cls(actor, jid)
        job._send_touch = False
        job.set_job_creator(job_creator)
        job.set_job_owner(job_owner)
        job.set_job_type(job_type)
        job.set_job_label(job_label)
        job.set_job_params(job_params)
        job.set_pool(pool)
        job.set_priority(priority)
        job.set_paused(paused)
        job._send_touch = True
        actor.send_job_created(jid)

        return job

    @classmethod
    def delete(cls, actor, jid, forced):
        try:
            job = cls(actor, jid)
        except ValueError:
            print('WARNING: Job.delete(): job not found.')
            return False

        if job.get_in_progress() and not forced:
            return False

        job.remove_from_pool_queue()

        actor._db.srem(actor.jobids_set_name, jid)
        nb = actor._db.delete(actor.job_hashes_namespace+'/'+jid)
        if not nb:
            print('WARNING: Job.delete(): did not delete job hash :/')
        return bool(nb)

    @classmethod
    def UnknownJobDict(cls, jid):
        return dict(
            jid=jid,
            label='!!!',
            created_on=0,
            creator='!!!',
            owner='!!!',
            job_type='!!!',
            job_params=((),{}),
            pool='!!!',
            paused=False,
            priority=0,
            in_progress=False,
            done=False,
            status='Unknown',
            icon_ref=('icons.status', 'WARN'),
        )

    def __init__(self, actor, jid):
        super(Job, self).__init__()
        self._actor = actor
        self._jid = jid
        self._send_touch = True

        self.assert_exists()

    def assert_exists(self):
        actor = self._actor
        if not actor._db.exists(actor.job_hashes_namespace+'/'+self._jid):
            raise ValueError('This job does not exists: {}'.format(self._jid))

    @contextlib.contextmanager
    def after_touch(self):
        prev = self._send_touch
        self._send_touch = False
        try:
            yield
        finally:
            self._send_touch = prev
            self.touched()

    def _set_field(self, name, dumps_value):
        self._actor._db.hset(
            self._actor.job_hashes_namespace+'/'+self._jid,
            name,
            dumps_value
        )

    def _get_field(self, name):
        return self._actor._db.hget(
            self._actor.job_hashes_namespace+'/'+self._jid,
            name
        )

    def to_dict(self):
        d = self._actor._db.hgetall(
            self._actor.job_hashes_namespace+'/'+self._jid,
        )
        d['created_on'] = float(d['created_on'])
        d['paused'] = d.get('paused', '1') == '1'
        d['priority'] = int(d.get('priority',0))

        d['in_progress'] = bool(d.get('in_progress', ''))
        d['done'] = bool(d.get('done', ''))

        d['jid'] = self._jid
        status, icon_ref = self.get_status_and_icon()
        d['status'] = status
        d['icon_ref'] = icon_ref

        d['has_error'] = self.get_has_error()
        if d['has_error']:
            d['error_message'] = self.get_error()
        return d 

    def get_status_and_icon(self):
        status = ''
        if self.get_has_error():
            status = 'Warn'
        elif self.get_paused():
            status = 'Inv'
        elif self.get_in_progress():
            status = 'WIP'
        elif self.get_done():
            status = 'Done'
        else:
            status = 'Wait'
        return status, ('icons.status', status.upper())

    def touched(self):
        if self._send_touch:
            self._actor.send_job_touched(self._jid)

    def jid(self):
        return self._jid

    def set_job_creator(self, login):
        self._set_field('creator', login)

    def get_job_creator(self):
        return self._get_field('creator')

    def set_job_owner(self, login):
        self._set_field('owner', login)

    def get_job_owner(self):
        return self._get_field('owner')

    def set_job_type(self, job_type):
        if self.get_in_progress():
            # too late:
            return
        self._set_field('job_type', job_type)
        self.touched()

    def get_job_type(self):
        return self._get_field('job_type')

    def set_job_label(self, job_label):
        self._set_field('job_label', job_label)
    
    def get_job_label(self):
        return self._get_field('job_label')

    def set_job_params(self, job_params):
        if self.get_in_progress():
            # too late:
            return
        self._set_field('job_params', json.dumps(job_params))
        self.touched()

    def get_job_params(self):
        args, kwargs = json.loads(self._get_field('job_params'))
        str_key_kwargs = {}
        for k, v in kwargs.items():
            str_key_kwargs[str(k)] = v
        return args, str_key_kwargs

    def add_to_pool_queue(self):
        q = self._actor.get_pool_queue(self.get_pool())
        score = float(self.get_priority())
        
        if self._actor.redis_version[0] < 3:
            self._actor._db.zadd(self._key(key), **{member: score})
        else:
            self._actor._db.zadd(
                q, mapping={self._jid: score}
            )
        
        self.touched()

    def remove_from_pool_queue(self):
        q = self._actor.get_pool_queue(self.get_pool())
        removed = self._actor._db.zrem(q, self._jid)
        if removed:
            self.touched()

    def set_oid(self, oid):
        if self.get_in_progress():
            # too late:
            return
        self._set_field('oid', str(oid))
        self.touched()

    def get_oid(self):
        '''
        Returns the flow Ojbect id that the workers will "execute"
        '''
        return self._get_field('oid')

    def set_pool(self, pool):
        if self.get_in_progress():
            # too late...
            return 

        if self.get_pool() is None:
            # first set is straight forward:
            self._set_field('pool', str(pool))
            self.touched()
            return 

        with self.after_touch():
            unpause = False
            if not self.get_paused():
                unpause = True
                self.set_paused(True)
            
            self._set_field('pool', str(pool))

            if unpause:
                self.set_paused(False)

    def get_pool(self):
        return self._get_field('pool')

    def set_paused(self, b):
        if self.get_in_progress():
            # too late
            return 

        if self.get_paused() == b:
            return 

        with self.after_touch():
            if not b:
                self.add_to_pool_queue()
            else:
                self.remove_from_pool_queue()
            self._set_field('paused', b and '1' or '')

    def get_paused(self):
        s = self._get_field('paused')
        return s == '1'

    def set_priority(self, priority):
        with self.after_touch():
            unpause = False
            if not self.get_paused():
                unpause = True
                self.set_paused(True)

            self._set_field('priority', str(priority))

            if unpause:
                self.set_paused(False)

    def get_priority(self):
        return int(self._get_field('priority') or 0)

    def set_in_progress(self):
        with self.after_touch():
            self._set_field('in_progress', '1')
            self._set_field('done', '')

    def get_in_progress(self):
        return bool(self._get_field('in_progress'))

    def set_node(self, node_id):
        with self.after_touch():
            self.remove_from_pool_queue()
            self._set_field('node', node_id)
            self._set_field('in_progress', '1')

    def set_done(self):
        with self.after_touch():
            self._set_field('in_progress', '')
            self._set_field('done', '1')

    def get_done(self):
        return bool(self._get_field('done'))

    def clear_error(self):
        with self.after_touch():
            self._set_field('has_error', '')
            self._set_field('error', '')

    def set_error(self, error_message):
        with self.after_touch():
            self._set_field('has_error', '1')
            self._set_field('error', json.dumps(error_message))

    def get_has_error(self):
        return bool(self._get_field('has_error'))

    def get_error(self):
        return json.loads(self._get_field('error'))

class WriteTee(object):
    '''
    Utility class used by LogTee
    '''
    def __init__(self, fh, orig, prefix=''):
        super(WriteTee, self).__init__()
        self._fh = fh
        self._orig = orig
        self._prefix = prefix

    def write(self, text):
        text = text.replace('\n', '\n'+self._prefix)
        self._fh.write(text)
        self._fh.flush()
        self._orig.write(text)

class LogTee(object):
    '''
    Utility class used by worker_execute_job to tee
    stdout and stderr to the log file.
    ''' 
    def __init__(self):
        self.log_filename = None
        self.log_fh = None
        
        self.orig_stdout = sys.stdout
        self.orig_stderr = sys.stderr
        
        self.stdout = None
        self.stderr = None
        
    def redirect(self, filename):
        dir = os.path.dirname(filename)
        if not os.path.exists(dir):
            os.makedirs(dir)
    
        self.log_filename = filename
        self.log_fh = open(self.log_filename, 'w')
    
        self.stdout = WriteTee(self.log_fh, self.orig_stdout)
        sys.stdout = self.stdout
    
        self.stderr = WriteTee(self.log_fh, self.orig_stderr, 'ERROR: ')
        sys.stderr = self.stderr
    
    def close_log(self):
        sys.stdout = self.orig_stdout
        self.stdout = None
    
        sys.stderr = self.orig_stderr
        self.stderr = None
    
        self.log_fh.close()
        self.log_fh = None

class Jobs(Actor):

    JOB_OID_ENVVAR_NAME = 'KABARET_JOBS_FLOW_OID'


    def __init__(self, session, jobs_db_index=9):
        super(Jobs, self).__init__(session)
        import redis
        self.redis_version = [int(i) for i in redis.__version__.split('.')]

        self._db = None
        self._jobs_db_index = jobs_db_index
        self._job_touched_unsubscribe = None
        self._joblist_touched_unsubscribe = None

        self.jobids_set_name = None
        self.job_hashes_namespace = None
        self.pool_queue_namespace = None

        self._node_is_alive = True

    def _create_cmds(self):
        return JobsCmds(self)

    def on_session_connected(self):
        self.log('Configuring Job Manager')
        cluster = self.session().get_actor('Cluster')

        cluster_name = cluster.get_cluster_name()
        self.jobids_set_name = '{}:Jobs:job_ids'.format(cluster_name)
        self.job_hashes_namespace = '{}:Jobs:jobs'.format(cluster_name)
        self.pool_queue_namespace = '{}:Jobs:pool_q'.format(cluster_name)

        self._db = cluster.get_db()

        self.log('Subcribing to jobs messages.')
        self._job_touched_unsubscribe = self.session().channels_subscribe(
            job_touched=self._on_job_touched
        )
        self._job_created_unsubscribe = self.session().channels_subscribe(
            job_created=self._on_job_created
        )
        self._job_deleted_unsubscribe = self.session().channels_subscribe(
            job_deleted=self._on_job_deleted
        )
        self._joblist_touched_unsubscribe = self.session().channels_subscribe(
            joblist_touched=self._on_joblist_touched
        )

    def die(self):
        if self._job_touched_unsubscribe is not None:
            self._job_touched_unsubscribe()

        if self._joblist_touched_unsubscribe is not None:
            self._joblist_touched_unsubscribe()

    def _on_job_touched(self, message):
        jid = message['data']
        self.session().dispatch_event('job_touched', jid=jid)

    def _on_job_created(self, message):
        jid = message['data']
        self.session().dispatch_event('job_created', jid=jid)

    def _on_job_deleted(self, message):
        jid = message['data']
        self.session().dispatch_event('job_deleted', jid=jid)

    def _on_joblist_touched(self, message):
        self.session().dispatch_event('joblist_touched')

    def send_job_touched(self, jid):
        self.session().publish(job_touched=jid)
    
    def send_job_created(self, jid):
        self.session().publish(job_created=jid)

    def send_job_deleted(self, jid):
        self.session().publish(job_deleted=jid)

    def send_joblist_touched(self):
        self.session().publish(joblist_touched=self.jobids_set_name)

    def get_pool_queue(self, pool_name):
        return self.pool_queue_namespace+':'+pool_name

    def get_current_login(self):
        return getpass.getuser()

    def create_job(self, job_type, job_label, job_creator, job_owner, job_params, pool, priority, paused):
        job = Job.create(self, job_type, job_label, job_creator, job_owner, job_params, pool, priority, paused)
        return job

    def get_all_jobs(self):
        jids = self._db.smembers(self.jobids_set_name)
        return [ self.get_job(jid) for jid in jids ]

    def get_job(self, jid):
        return Job(self, jid)

    def delete_job(self, jid, forced):
        '''
        Delete the given job if it is not running.
        If forced is True, deletes the job whatever its status.
    
        Returns True if the job was actually deleted (found and allowed for deletion)
        '''
        deleted = Job.delete(self, jid, forced)
        if deleted:
            self.send_job_deleted(jid)
        return deleted


    #
    #   NODE STUFF
    #
    '''
In Kabaret Script View, access this actor with:

    self.root().session().get_actor('Jobs')

Or use the cmds like:

# JOB CREATE / LIST / DELTE
jid = self.root().session().cmds.Jobs.create_flow_job(
    oid='/path/to/my/job_object', pool='test', priority=100, paused=True
)

self.root().session().cmds.Jobs.set_job_paused(jid, False)
self.root().session().cmds.Jobs.delete_job(jid)

jobs = self.root().session().cmds.Jobs.list_jobs()
for job in jobs:
    self.root().session().cmds.Jobs.delete_job(job['jid'])

# NODE CONTROL

jid = self.root().session().cmds.Jobs.kill_node_job(
    node_id='<NODE_ID>'
)

jid = self.root().session().cmds.Jobs.stop_node(
    node_id='<NODE_ID>'
)

    '''
    def stop_node(self):
        #TODO: we should send a node list touched message 
        self._node_is_alive = False

    def node_stopped(self):
        return not self._node_is_alive

    def _poll_queues(self, qs, node_id):
        jids = None
        for q in qs:
            jids = self._db.zrange(q, 0, 0)
            if jids:
                break

        if jids:
            job = self.get_job(jids[0])
            job.set_node(node_id)
            return job

        return None

    def poll_control(self, node_id):
        qs = [
            self.get_pool_queue(node_id)
        ]
        return self._poll_queues(qs, node_id)

    def poll_pools(self, pools, node_id):
        qs = [
            self.get_pool_queue(pool)
            for pool in pools+[node_id]
        ]
        return self._poll_queues(qs, node_id)

    def node_do_control_job(self, cmd, *args, **kwargs):
        if cmd == 'kill_job':
            print('Killing Job')
            print('     !!! not yet implemented !!!')
        
        elif cmd == 'stop':
            self.stop_node()

        else:
            raise ValueError('Unknown Node control cmd {!r}'.format(cmd))

    def node_spawn_flow_worker(self, oid, interpretor=None, worker_filename=None, show_console=True):
        '''
        Will create a subprocess that runs the given worker_filename (or default) with
        the given python interpretor (or current).

        The worker_filename must be a script which executes the job.
        The default one (kabaret.jobs.jobs_flow_worker) uses the env to connect
        to the same cluster as this one and finds the oid to execute in env.
        '''
        print('Spawning Worker to execute on oid={!r}'.format(oid))
        interpretor = interpretor or sys.executable

        if worker_filename is None:
            import libreflow.utils.kabaret.jobs.jobs_flow_worker as worker
            worker_filename = worker.__file__

        env = os.environ.copy()
        env[self.JOB_OID_ENVVAR_NAME] = str(oid) # json dump in redis made it a unicode :/

        # using -u for unbuffered output. Don't really know if useful :p yolo ^_^ 
        cmd = [interpretor, '-u', worker_filename]

        shell = False
        if show_console:
            if platform.system() == 'Windows':
                cmd = ['start', '/WAIT', 'cmd', '/C']+cmd
                shell = True
            else:
                print("!!! Warning: show_console option ignored for system other that Windows :/")

        try:
            popen = subprocess.Popen(cmd, env=env, shell=shell)
        except:
            traceback.print_exc()
            print(
                '!!! Error Spawning Worker Subprocess:\n#--- CMD was:\n{}\n#--- ENV was:\n{}\n'.format(
                    ' '.join(cmd),
                    '\n'.join([ '{}:{!r}'.format(k,v) for k, v in env.iteritems() ])
                )
            )
            raise
        
        return popen


    def node_execute_job(self, job):
        job_type = job.get_job_type()
        args, kwargs = job.get_job_params()
        print(
            'Executing {} job: {}, {} ({})'.format(
                job_type, args, kwargs, job.jid()
            )
        )

        if job_type == 'NODE_CMD':
            try:
                job.set_in_progress()
                self.node_do_control_job(*args, **kwargs)
            except:
                error_message = traceback.format_exc()
                print('------------------------')
                print(error_message)
                print('------------------------')
                print('ERROR EXECUTING NODE CMD')
                job.set_error('ERROR EXECUTING NODE CMD:\n'+error_message)
                return None
            else:
                print('Command done')
                job.set_done()
                return None

        elif job_type == 'flow':
            try:
                popen = self.node_spawn_flow_worker(*args, **kwargs)
            except:
                error_message = traceback.format_exc()
                print('------------------------')
                print(error_message)
                print('------------------------')
                print('ERROR SPAWING FLOW WORKER')
                job.set_error('ERROR SPAWING FLOW WORKER:\n'+error_message)
                return None
            else:
                print('Worker Spawned.')
                return popen

    #
    #   WORKER STUFF
    #

    def worker_execute_job(self):
        '''
        This is called by the default JobsWorkerSession to execute
        a job picked by the JobsNodeSession which spawned it.
        '''
        try:
            job_oid = os.environ[self.JOB_OID_ENVVAR_NAME]
        except KeyError:
            raise Exception('Could not find {} env-var :/'.format(self.JOB_OID_ENVVAR_NAME))

        import kabaret.flow as flow
        try:
            job = self.session().get_actor('Flow').get_object(job_oid)
        except (flow.MissingChildError, flow.MissingRelationError):
            raise Exception('Could not find job to execute: {}'.format(job_oid))

        log_filename = job.get_log_filename()
        log_tee = LogTee()
        log_tee.redirect(log_filename)
        try:
            job.execute()
        finally:
            log_tee.close_log()



from qtpy import QtCore

from kabaret.app.ui import gui
from kabaret.app.session import KabaretSession
from kabaret.script_view import ScriptView
from kabaret.subprocess_manager import SubprocessManager, LauncherToolBar

from libreflow.utils.kabaret.jobs.jobs_node import JobsNodeSession

from .utils.kabaret.jobs.jobs_actor import Jobs


class BaseGUISession(gui.KabaretStandaloneGUISession):
    def register_view_types(self):
        super(BaseGUISession, self).register_view_types()

    def _create_actors(self):
        super(BaseGUISession, self)._create_actors()

        # Configure SubprocessManager
        subprocess_manager = SubprocessManager(self)


class DebugGUISession(BaseGUISession):
    def register_view_types(self):
        super(DebugGUISession, self).register_view_types()

        type_name = self.register_view_type(ScriptView)
        self.add_view(type_name, hidden=True, area=QtCore.Qt.RightDockWidgetArea)


class BaseCLISession(KabaretSession):
    def _create_actors(self):
        super(BaseCLISession, self)._create_actors()

        # Configure SubprocessManager
        SubprocessManager(self)
        Jobs(self)


class JobsNodeSession(JobsNodeSession):
    pass



from __future__ import print_function

import six
import time
import json
import os

from qtpy import QtWidgets, QtCore

try:
    from kabaret.app.ui.gui.widgets.widget_view import DockedView
except ImportError:
    raise RuntimeError('Your kabaret version is too old, for the JobView. Please update !')

from kabaret.app import resources

class JobListParamItem(QtWidgets.QTreeWidgetItem):

    def __init__(self, job_item, v, k=None):
        super(JobListParamItem, self).__init__(job_item)
        self._job_item = job_item
        self.k = k
        self.v = v 

        if k is not None:
            self.setText(0, '{}={}'.format(k, v))
        else:
            self.setText(0, v)

        self.setIcon(0, resources.get_icon(('icons.gui', 'tag-black-shape')))

    def job_item(self):
        return self._job_item

class JobListItem(QtWidgets.QTreeWidgetItem):

    def __init__(self, tree, job):
        super(JobListItem, self).__init__(tree)
        self.job = None
        self.args = ()
        self.kwargs = {}
        self._match_str = ''
        self.set_job(job)

    def job_item(self):
        return self

    def set_job(self, job):
        nb = self.childCount()
        for i in range(nb):
            child = self.takeChild(i)
            # Destroy it ?

        self.job = job
        self.args = ()
        self.kwargs = {}
        self._match_str = ''
        self._update()

    def _update(self):
        j = self.job
        self.setText(1, j['job_type'])
        self.setText(2, str(j['job_params']))
        self.setText(3, j['job_label'])
        self.setText(4, j['owner'])
        self.setText(5, j['pool'])
        self.setText(6, str(j['priority']))
        self.setText(7, j.get('node', ''))

        # self.setText(7, str(j['paused']))
        # self.setText(8, str(j['in_progress']))
        # self.setText(9, str(j['done']))

        self.setText(8, j['creator'])
        
        ts = j['created_on']
        date_str = time.ctime(ts)

        self.setText(9, date_str)
        self.setText(10, str(ts))
        self.setText(11, j['jid'])

        status = j['status']
        self.setIcon(0, resources.get_icon(j['icon_ref']))
        self.setText(0, '')
        
        matchers = [
            j['jid'],
            j['owner'], status, j['job_type'],
            j['pool'],
            j['creator'], date_str
        ]
        try:
            self.args, self.kwargs = json.loads(str(j['job_params']))
        except json.JSONDecodeError:
            pass
        else:
            for v in self.args:
                child = JobListParamItem(self, v)
                matchers.append(v)

            for k, v in self.kwargs.items():
                child = JobListParamItem(self, v, k)
                matchers.append('{}={}'.format(k,v))

        self._match_str = '^'.join(matchers)

    def jid(self):
        return self.job['jid']

    def paused(self):
        return self.job['paused']

    def in_progress(self):
        return self.job['in_progress']

    def done(self):
        return self.job['done']

    def matches(self, filter):
        return filter in self._match_str


class JobListHeaderView(QtWidgets.QHeaderView):

    def __init__(self, parent, orientation=QtCore.Qt.Horizontal):
        super(JobListHeaderView, self).__init__(orientation, parent)

        self.setDefaultAlignment(QtCore.Qt.AlignLeft)
        self.setSectionsMovable(True)

        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_column_selection_requested)

        self._popup_menu = QtWidgets.QMenu(self)

        column_names, columns_visible = self.get_columns()
        for col_i, (col_name, col_visible) in enumerate(zip(column_names, columns_visible)):
            a = self._popup_menu.addAction(col_name, lambda i=col_i: self._on_column_visible_toggled(i))
            a.setCheckable(True)
            a.setChecked(col_visible)

    def get_columns(self):
        names = (
            'Status',
            'Type', 'Params', 'Label',
            'Owner', 
            'Pool', 'Priority', 
            'Node', 

            # 'Paused', 'In Progress', 'Done', 
            
            'Creator',
            'Created On', 

            'TIMESTAMP', 'ID'
        )
        visible = (
            True,                   # Status
            False,  False, True,    # Type, params, label
            False, False, False,    # Owner, pool, priority
            True,                   # Node
            # False, False, False,    # Paused, In progress, Done
            True,  True,            # Creator, created on
            False, False            # Timestamp, ID
        )

        return names, visible
    
    def _on_column_selection_requested(self, pos):
        self._popup_menu.popup(self.viewport().mapToGlobal(pos))
    
    def _on_column_visible_toggled(self, column):
        hidden = self.isSectionHidden(column)

        if column == 3:
            self.setStretchLastSection(not hidden)
        
        self.setSectionHidden(column, not hidden)


class JobList(QtWidgets.QTreeWidget):

    def __init__(self, parent, session):
        super(JobList, self).__init__(parent)
        self.session = session

        self.setHeader(JobListHeaderView(self))
        column_names, columns_visible = self.header().get_columns()
        self.setHeaderLabels(column_names)

        self.header().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Interactive)
        self.header().setSectionResizeMode(3, QtWidgets.QHeaderView.Stretch)
        
        for col_i, col_visible in enumerate(columns_visible):
            self.setColumnHidden(col_i, not col_visible)

        self.setSortingEnabled(True)
        self.sortByColumn(len(column_names)-2, QtCore.Qt.DescendingOrder)

        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_popup_menu_request)

        self._popup_menu = QtWidgets.QMenu(self)

        self._filter = None
        self._jid_to_item = {}

    def set_item_filter(self, filter):
        self._filter = filter
        self.apply_filter()

    def refresh(self):
        #TODO: intelligent refresh: remove deleted jids, add created jids
        self._jid_to_item.clear()
        self.clear()
        for job in self.session.cmds.Jobs.list_jobs():
            item = JobListItem(self, job)
            self._jid_to_item[item.jid()] = item
            if self._filter and not item.matches(self._filter):
                item.setHidden(True)

    def apply_filter(self):
        root = self.invisibleRootItem()
        nb = root.childCount()
        for i in range(nb):
            item = root.child(i)
            if not self._filter:
                item.setHidden(False)
            else:
                item.setHidden(
                    not item.matches(self._filter)
                )

    def refresh_job(self, jid):
        item = self._jid_to_item.get(jid)
        if item is not None:
            job = self.session.cmds.Jobs.get_job_info(jid)
            item.set_job(job)
    
    def create_job(self, jid):
        job = self.session.cmds.Jobs.get_job_info(jid)
        item = JobListItem(self, job)
        self._jid_to_item[jid] = item
        if self._filter and not item.matches(self._filter):
            item.setHidden(True)
    
    def delete_job(self, jid):
        job_index = self.indexOfTopLevelItem(self._jid_to_item[jid])
        self.takeTopLevelItem(job_index)

    def _on_popup_menu_request(self, pos):
        item = self.itemAt(pos)
        if item is None:
            m = self._popup_menu
            m.clear()
            m.addAction('Refresh', self.refresh)

        else:
            item = item.job_item()
            m = self._popup_menu
            m.clear()
            if item.paused():
                m.addAction('Un-Pause', lambda item=item: self._unpause(item))
            if 'oid' in item.kwargs:
                oid = item.kwargs['oid']
                name = os.path.basename(oid)
                m.addAction('Goto "{}"'.format(name), lambda item=item: self._goto(item))
            m.addAction('Delete', lambda item=item: self._delete(item))

        self._popup_menu.popup(self.viewport().mapToGlobal(pos))

    def _unpause(self, item):
        self.session.cmds.Jobs.set_job_paused(item.jid(), False)

    def _delete(self, item):
        self.session.cmds.Jobs.delete_job(item.jid())

    def _goto(self, item):
        if not self.session.is_gui():
            return

        oid = str(item.kwargs['oid'])
        view = self.session.add_view('Flow', area=None, oid=oid)
        # ensure visible if tabbed:
        view.show()

class JobsView(DockedView):

    @classmethod
    def view_type_name(cls):
        return 'Jobs'

    def __init__(self, *args, **kwargs):
        super(JobsView, self).__init__(*args, **kwargs)

    def _build(self, top_parent, top_layout, main_parent, header_parent, header_layout):
        self.add_header_tool('*', '*', 'Duplicate View', self.duplicate_view)

        self._filter_le = QtWidgets.QLineEdit(main_parent)
        self._filter_le.setPlaceholderText('Filter Jobs...')
        self._filter_le.textChanged.connect(self._on_filter_change)

        self._job_list = JobList(main_parent, self.session)
        self._job_list_refreshed = False

        # Set job filter to default if defined
        # in environment (may be at session startup)
        if "JOBS_DEFAULT_FILTER" in os.environ:
            self._filter_le.setText(os.environ["JOBS_DEFAULT_FILTER"] + '_')

        lo = QtWidgets.QVBoxLayout()
        lo.setContentsMargins(0, 0, 0, 0)
        lo.setSpacing(0)
        lo.addWidget(self._filter_le)
        lo.addWidget(self._job_list)

        main_parent.setLayout(lo)


        self.view_menu.setTitle('Jobs')

        a = self.view_menu.addAction('Refresh', self.refresh)
        self.view_menu.addSeparator()
        a = self.view_menu.addAction('Create Test Job', self.create_test_job)

    def _on_filter_change(self):
        self._job_list.set_item_filter(self._filter_le.text())

    def refresh(self):
        self._job_list.refresh()

    def create_test_job(self):
        jid = self.session.cmds.Jobs.create_flow_job(
            oid='/project/path/to/the/job_object',
            pool='test', 
            priority=50,
            paused=True,
        )
        print('NEW JOB ID:', jid)

    def on_show(self):
        if not self._job_list_refreshed:
            self.refresh()
            self._job_list_refreshed = True

    def receive_event(self, event, data):
        # Is it really a good thing to update job list items in background,
        # even if the view is not shown ?
        if event == 'joblist_touched':
            if self.isVisible():
                self._job_list.refresh() 

        elif event == 'job_touched':
            jid = data['jid']
            self._job_list.refresh_job(jid)
        elif event == 'job_created':
            jid = data['jid']
            self._job_list.create_job(jid)
        elif event == 'job_deleted':
            jid = data['jid']
            self._job_list.delete_job(jid)


import os

from kabaret import flow
from kabaret.flow_contextual_dict import ContextualView, get_contextual_dict

from libreflow import baseflow
from .departments import Department


class AssetDepartments(flow.Object):
    design = flow.Child(Department).ui(expanded=True)
    cycle = flow.Child(Department).ui(expanded=True)


class Asset(baseflow.lib.Asset):

    departments = flow.Child(AssetDepartments).ui(expanded=True)

    def compute_child_value(self, child_value):
        if child_value is self.kitsu_url:
            child_value.set(
                "%s/episodes/all/assets/%s"
                % (self.root().project().kitsu_url.get(), self.kitsu_id.get())
            )


class Assets(baseflow.lib.Assets):

    _asset_type = flow.Parent()

    create_asset = flow.Child(baseflow.maputils.SimpleCreateAction)
    clear_map = flow.Child(baseflow.maputils.ClearMapAction).ui(hidden=True)
    with flow.group("Kitsu"):
        update_kitsu_settings = flow.Child(baseflow.film.UpdateItemsKitsuSettings)

    @classmethod
    def mapped_type(cls):
        return Asset

    def columns(self):
        return ["Asset"]

    def _fill_row_cells(self, row, item):
        row["Asset"] = item.name()

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return self._asset_type.get_default_contextual_edits(context_name)


class AssetType(flow.Object):

    _asset_types = flow.Parent()
    assets = flow.Child(Assets).ui(expanded=True)
    settings = flow.Child(ContextualView).ui(hidden=True)

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(asset_type=self.name())


class AssetTypes(flow.Map):

    create_asset_type = flow.Child(baseflow.maputils.SimpleCreateAction)
    clear_asset_types = flow.Child(baseflow.maputils.ClearMapAction)

    @classmethod
    def mapped_type(cls):
        return AssetType


class AssetLib(flow.Object):

    asset_types = flow.Child(AssetTypes).ui(expanded=True)

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(file_category="LIB")



import os
import sys
import platform
import subprocess
from pathlib import Path
from datetime import datetime

from kabaret import flow
from kabaret.subprocess_manager.runner_factory import Runner, logger
# from kabaret.subprocess_manager.flow import RunAction

from ..utils.kabaret.subprocess_manager.runner_factory import Runner
from ..utils.kabaret.subprocess_manager.flow import RunAction
from ..resources.icons import flow as flow_icons


CHOICES = ["blend", "kra", "png", "jpg", "txt", "abc", "mov", "psd", "aep", "zip", "mp4", "fbx", "ai", "json", "jsx", "obj", "wav"]
CHOICES_ICONS = {
            "blend": ("icons.libreflow", "blender"),
            "kra": ("icons.libreflow", "krita"),
            "png": ("icons.gui", "picture"),
            "jpg": ("icons.gui", "picture"),
            "txt": ('icons.gui', 'text-file-1'),
            "abc": ("icons.flow", "alembic"),
            "aep": ("icons.libreflow", "afterfx"),
            "psd": ("icons.flow", "photoshop"),
            "mov": ("icons.flow", "quicktime"),
            "ai":  ("icons.libreflow", "illustrator"),
            "zip": ("icons.libreflow", "archive"),
            "mp4": ("icons.gui", "youtube-logo"),
            "fbx": ("icons.libreflow", "fbx"),
            "json": ("icons.libreflow", "json"),
            "jsx": ("icons.libreflow", "jsx"),
            "obj": ("icons.libreflow", "3d-object"),
            "wav": ("icons.gui", "youtube-logo"),
        }



# Runners
# -----------------


class DefaultEditor(Runner):
    @classmethod
    def can_edit(cls, filename):
        return True

    def executable(self):
        if platform.system() == "Darwin":
            return "open"
        elif platform.system() == "Linux":
            return "xdg-open"
        return None

    def run(self):
        if platform.system() == "Windows":
            os.startfile(self.argv()[0])
        else:
            super(DefaultEditor, self).run()


class EditFileRunner(Runner):

    ICON = ("icons.flow", "action")

    @classmethod
    def can_edit(cls, filename):
        ext = os.path.splitext(filename)[1]
        supported_exts = cls.supported_extensions()

        return not bool(supported_exts) or ext in supported_exts

    @classmethod
    def supported_extensions(cls):
        """
        Supported file extensions.

        Return None by default to allow any extension.
        """
        return None

    def show_terminal(self):
        return False

    def exec_env_var(self, version=None):
        key = self.__class__.__name__.upper()

        if version is not None:
            key += '_%s' % version.upper().replace('.', '_')
        
        return '%s_EXEC_PATH' % key

    def executable(self):
        try:
            exec_path = os.environ[self.exec_env_var(self.version)]
            print('[RUNNER] exec path: ', exec_path)
            return exec_path
        except KeyError:
            exec_path = os.environ[self.exec_env_var()]
            print('[RUNNER] exec path: ', exec_path)
            return exec_path


class ImageMagick(EditFileRunner):

    @classmethod
    def can_edit(cls, filename):
        return True


class Blender(EditFileRunner):

    ICON = ("icons.libreflow", "blender")
    TAGS = [
        "Modeling",
        "Sculpting",
        "Animation",
        "Rigging",
        "3D Drawing",
        "Rendering",
        "Simulation",
        "Video Editing",
        "VFX",
    ]

    @classmethod
    def supported_versions(cls):
        return ["2.83", "2.90", "2.91", "2.92", "2.93"]

    @classmethod
    def supported_extensions(cls):
        return [".blend"]


class Krita(EditFileRunner):

    ICON = ("icons.libreflow", "krita")
    TAGS = ["2D Drawing", "Image Editing"]

    @classmethod
    def supported_versions(cls):
        return ["4.3.0"]

    @classmethod
    def supported_extensions(cls):
        return [".kra", ".png", ".jpg"]


class AfterEffects(EditFileRunner):

    ICON = ("icons.libreflow", "afterfx")

    @classmethod
    def supported_extensions(cls):
        return [".aep", ".png", ".jpg"]


class AfterEffectsRender(EditFileRunner):

    ICON = ("icons.libreflow", "afterfx")

    @classmethod
    def supported_extensions(cls):
        return [".aep"]


class VSCodium(EditFileRunner):

    ICON = ("icons.libreflow", "vscodium")
    TAGS = ["Text editing", "IDE"]

    @classmethod
    def supported_extensions(cls):
        return [".txt"]


class NotepadPP(EditFileRunner):

    ICON = ("icons.flow", "notepad")
    TAGS = ["Text editing"]

    @classmethod
    def supported_extensions(cls):
        return [".txt"]


class Firefox(EditFileRunner):

    ICON = ("icons.flow", "notepad")
    TAGS = ["Browser"]

    @classmethod
    def can_edit(cls, filename):
        return True


class RV(EditFileRunner):

    ICON = ("icons.applications", "rv")
    TAGS = ['Video editing']

    def show_terminal(self):
        return True


class PythonRunner(Runner):

    def executable(self):
        return sys.executable
    
    def show_terminal(self):
        return False

    def keep_terminal(self):
        return False


class MarkSequenceRunner(PythonRunner):
    
    TAGS = ['Mark image sequence']

    def argv(self):
        script_path = "%s/../scripts/mark_sequence.py" % os.path.dirname(__file__)
        return [script_path] + self.extra_argv


class SessionWorker(PythonRunner):

    def argv(self):
        args = [
            "%s/../scripts/session_worker.py" % (
                os.path.dirname(__file__)
            ),
            self.runner_name()
        ]
        args += self.extra_argv
        return args


class LaunchSessionWorker(RunAction):

    def runner_name_and_tags(self):
        return "SessionWorker", []
    
    def launcher_oid(self):
        raise NotImplementedError()

    def launcher_exec_func_name(self):
        raise NotImplementedError()
    
    def extra_argv(self):
        return [
            self.launcher_oid(),
            self.launcher_exec_func_name()
        ]
    
    def run(self, button):
        '''
        Sets the environment variable which contains the runner executable path
        before launching the runner.
        '''
        name, tags = self.runner_name_and_tags()
        
        rid = self.root().session().cmds.SubprocessManager.run(
            runner_name=name,
            tags=tags,
            version=self.get_version(button),
            label=self.get_run_label(),
            extra_argv=self.extra_argv(),
            extra_env=self.extra_env(),
        )
        return self.get_result(runner_id=rid)


class RunnerChoiceValue(flow.values.ChoiceValue):

    _extension = flow.Parent(2)

    def choices(self):
        factory = self.root().project().get_factory()

        runner_names = [
            nat[0]
            for nat in factory.find_runners("*.{}".format(self._extension.name()))
        ]

        return runner_names


class RunnerVersionChoiceValue(flow.values.ChoiceValue):

    STRICT_CHOICES = False

    _action = flow.Parent()

    def choices(self):
        factory = self.root().project().get_factory()
        runner_name = self._action.application.get()
        runner_versions = factory.get_runner_versions(runner_name)

        if not runner_versions:
            runner_versions = []

        return ['default'] + runner_versions


class ChangeDefaultRunner(flow.Action):

    _extension = flow.Parent()

    application = flow.Param(None, RunnerChoiceValue).watched()
    version = flow.Param(None, RunnerVersionChoiceValue)

    def get_buttons(self):
        return ["Validate", "Cancel"]

    def child_value_changed(self, child_value):
        if child_value is self.application:
            version_choices = self.version.choices()
            if not version_choices:
                self.version.set("")
            else:
                self.version.set(version_choices[0])

            self.version.touch()

    def run(self, button):
        if button == "Cancel":
            return

        factory = self.root().project().get_factory()
        runner_names = [nat[0] for nat in factory.find_runners()]

        if not self.application.get() in runner_names:
            self.root().session().log_debug("Application not recognized")
            return self.get_result(close=False)

        self._extension.runner_name.set(self.application.get())
        self._extension.runner_version.set(self.version.get())
        self._extension.touch()


class DefaultRunner(flow.Object):

    runner_name = flow.Param("").ui(editable=False)
    runner_version = flow.Param("").ui(editable=False)

    change_default_application = flow.Child(ChangeDefaultRunner)

    def get_runner(self):
        factory = self.root().project().get_factory()

        if not self.runner_name.get():
            runners = factory.find_runners(
                edited_filename="*.{ext}".format(ext=self.name())
            )

            if runners:
                runner_name = runners[0][0]
                runner_version = factory.get_runner_versions(runner_name)[0] or ""
                self.runner_name.set(runner_name)
                self.runner_version.set(runner_version)

        return factory.get_runner(
            runner_name=self.runner_name.get(),
            tags=[],
            version=self.runner_version.get(),
        )


class DefaultRunners(flow.DynamicMap):
    @classmethod
    def mapped_type(cls):
        return DefaultRunner

    def mapped_names(self, page_num=0, page_size=None):
        return CHOICES

    def columns(self):
        return ["Extension", "Default application", "Version"]

    def _fill_row_cells(self, row, item):
        row["Extension"] = item.name()
        row["Default application"] = item.runner_name.get()
        row["Version"] = item.runner_version.get()

    def _fill_row_style(self, style, item, row):
        style["activate_oid"] = item.change_default_application.oid()

        runner = item.get_runner()
        if runner:
            style["icon"] = runner.runner_icon()
            return

        style["icon"] = ("icons.gui", "cog-wheel-silhouette")



import os
import json
import time
import re
from collections import defaultdict

from kabaret import flow

from .maputils import SimpleCreateAction


class DependencyTemplate(flow.Object):

    dependencies = flow.Param('').ui(editor='textarea')
    defaults = flow.OrderedStringSetParam()

    def get_dependencies(self):
        return json.loads(self.dependencies.get())

    def get_default_dependency_names(self):
        return self.defaults.get()

    def get_dependency_files(self, name):
        deps = self.get_dependencies()[name]
        
        if 'files' in deps:
            return deps['files']
        
        return deps

    def get_asset_files(self, asset_name):
        kitsu_bindings = self.root().project().kitsu_bindings()
        asset = kitsu_bindings.get_asset_data(asset_name)

        return self.get_dependency_files('kitsu.'+asset['type'])


class DependencyTemplates(flow.Map):

    add_template = flow.Child(SimpleCreateAction)

    @classmethod
    def mapped_type(cls):
        return DependencyTemplate


class FileSystemItem:

    def __init__(self, entity_name, entity_oid, department, name, revision, project):
        self._entity_name = entity_name
        self._entity_oid = entity_oid
        self._department = department
        self._name = name
        self._project = project
        self._is_real = False
        self._path = None

        self._revision = self._get_revision_name(revision)

    @staticmethod
    def _get_shot_entity_name(department, file_name):
        '''
        Cheat function, for now, to get shot entity name based on file name.
        '''
        return file_name.split('.')[0]

    @classmethod
    def _get_name(cls, flow_name):
        raise NotImplementedError()

    @classmethod
    def _args_from_path(cls, path, project):
        entity_oid = project.oid()
        entity_name = 'undefined'
        shot_match = re.search("/(siren|siren_test)/sq\d\d/sc\d\d\d\d/.*/.*/v\d\d\d", path)

        if shot_match is not None:
            subpath = shot_match.group(0)[1:]
            (
                film,
                sequence,
                shot,
                department,
                name,
                revision,
            ) = subpath.split('/')

            name = cls._get_name(name)
            entity_name = cls._get_shot_entity_name(department, name)

            entity_oid += f"/films/{film}/sequences/{sequence}/shots/{shot}"
            return entity_name, entity_oid, department, name, revision

        asset_match = re.search("/lib/(chars|props|sets)/.*/.*/.*/.*/v\d\d\d", path)

        if asset_match is not None:
            subpath = asset_match.group(0)[5:]
            (
                asset_type,
                asset_family,
                asset_name,
                department,
                name,
                revision,
            ) = subpath.split('/')

            name = cls._get_name(name)

            entity_oid += f"/asset_lib/asset_types/{asset_type}/asset_families/{asset_family}/assets/{asset_name}"
            return asset_name, entity_oid, department, name, revision

        return entity_name, entity_oid, 'undefined', 'undefined', 'undefined'

    def _get_revision_name(self, revision):
        name = revision

        if revision == 'last':
            obj = self.related_object()
            name = None

            if obj is not None:
                revision_object = obj.get_head_revision()

                if revision_object is not None:
                    name = revision_object.name()

        return name

    @classmethod
    def create_from_path(cls, path, project):
        raise NotImplementedError()

    def exists(self):
        session = self._project.root().session()
        return session.cmds.Flow.exists(self.oid())

    def related_object(self):
        if not self.exists():
            return None
        return self._project.root().get_object(self.oid())

    def name(self):
        return self._name

    def flow_name(self):
        return self._name.replace('.', '_')

    def oid(self):
        return (
              self.entity_oid()
            + "/departments/"
            + self._department
            + "/files/"
            + self.flow_name()
        )

    def last_revision(self):
        return self._get_revision_name('last')

    def revision_exists(self):
        if self.revision_oid() is None:
            return False

        session = self._project.root().session()
        return session.cmds.Flow.exists(self.revision_oid())

    def related_revision_object(self, revision_name=None):
        if not self.revision_exists():
            return None

        if revision_name is None:
            revision_name = self.revision()

        return self._project.root().get_object(self.revision_oid())

    def revision(self):
        return self._revision

    def revision_oid(self):
        if self.revision() is None:
            return None

        return self.oid() + "/history/revisions/" + self.revision()

    def entity_name(self):
        return self._entity_name

    def entity_oid(self):
        return self._entity_oid

    def sync_status(self, revision, site_name=None, exchange=False):
        return revision.get_sync_status(site_name, exchange)

    def source_site(self):
        if not self.revision_exists():
            return None

        revision = self.related_revision_object()
        return revision.site.get()

    def status(self, site_name=None):
        if not self.revision_exists():
            return 'NotAvailable'

        revision = self.related_revision_object()

        if self.sync_status(revision, site_name=site_name) == 'Available':
            return 'Available'
        elif self.sync_status(revision, site_name=site_name) == 'Requested':
            return 'Requested'
        elif self.sync_status(revision, exchange=True) == 'Available':
            return 'Downloadable'
        else:
            return 'Requestable'

    def is_real(self):
        return self._is_real

    def path(self):
        return self._path


class File(FileSystemItem):

    @classmethod
    def create_from_path(cls, path, project):
        path = path.replace('\\', '/')
        entity_name, entity_oid, dept, name, rev = cls._args_from_path(path, project)

        file = cls(entity_name, entity_oid, dept, name, rev, project)
        file._is_real = True
        file._path = path

        return file

    @classmethod
    def _get_name(cls, flow_name):
        extension_index = flow_name.rfind('_')
        return flow_name[:extension_index] + '.' + flow_name[extension_index+1:]


class Folder(FileSystemItem):

    @classmethod
    def create_from_path(cls, path, project):
        path = path.replace('\\', '/')
        entity_name, entity_oid, dept, name, rev = cls._args_from_path(path, project)
        name = name.split('_')[-1]

        folder = cls(entity_name, entity_oid, dept, name, rev, project)
        folder._is_real = True
        folder._path = path

        return folder

    @classmethod
    def _get_name(cls, flow_name):
        return flow_name


def get_real_files(paths, flow_project):
    folders = defaultdict(list)
    real_files = []

    for path in paths:
        path = os.path.normpath(path)
        folders[os.path.dirname(path)].append(os.path.basename(path))

    for folder, files in folders.items():
        files = list(set(files))
        
        if len(files) > 1:
            item = Folder.create_from_path(folder, flow_project)
        else:
            item = File.create_from_path(folder, flow_project)
            
            if not item.exists():
                item = Folder.create_from_path(folder, flow_project)
            
        real_files.append(item)

    return real_files


def resolve_files(predictive_files, real_files=None, site_name=None):
    resolved_files = []

    if real_files is None:
        for f in predictive_files:
            resolved_files.append({
                'entity_name': f.entity_name(),
                'entity_oid': f.entity_oid(),
                'file_name': f.name(),
                'revision': f.revision(),
                'file_oid': f.oid(),
                'revision_oid': f.revision_oid(),
                'status': f.status(site_name=site_name),
                'in_breakdown': True,
                'source_site': f.source_site(),
                'last_revision': f.last_revision(),
                'is_alien': False,
                'is_real': False,
                'path': f.path(),
            })

        return resolved_files

    # Sort predictive and real files according to the file oid
    files = defaultdict(lambda: defaultdict(list))

    for pred_file in predictive_files:
        files[pred_file.oid()]['pred'].append(pred_file)

    for real_file in real_files:
        files[real_file.oid()]['real'].append(real_file)

    # Resolve predictive and real files
    for file_oid in files:
        pred_and_real = files[file_oid]
        pred_files = pred_and_real['pred'] # Only one predictive file here
        result_files = pred_and_real['real']
        in_breakdown = bool(pred_files)

        if not result_files:
            result_files = pred_files

        for f in result_files:
            resolved_files.append({
                'entity_name': f.entity_name(),
                'entity_oid': f.entity_oid(),
                'file_name': f.name(),
                'revision': f.revision(),
                'file_oid': file_oid,
                'revision_oid': f.revision_oid(),
                'status': f.status(site_name=site_name),
                'in_breakdown': in_breakdown,
                'source_site': f.source_site(),
                'last_revision': f.last_revision(),
                'is_alien': not in_breakdown and not f.exists(),
                'is_real': f.is_real(),
                'path': f.path(),
            })

    return resolved_files


def get_dependencies(leaf, site_name=None, predictive=False, real=False, revision_name=None):
    project = leaf.root().project()

    predictive_files = []
    real_files = None

    # Get object's real dependencies
    if real:
        try:
            dependency_getter = getattr(leaf, 'get_real_dependencies')
        except AttributeError:
            pass
        else:
            real_file_paths = dependency_getter(revision_name)
            real_files = get_real_files(real_file_paths, project)

    # Get object's dependency template
    if predictive:
        templates = project.admin.dependency_templates
        template_name = leaf.name()
        kitsu_casting = list()

        # Check if template name is explicitly provided
        try:
            template_getter = getattr(leaf, 'get_dependency_template')
        except AttributeError:
            pass
        else:
            template_name, entity_oid, kitsu_casting = template_getter()

        if not template_name or not templates.has_mapped_name(template_name):
            leaf.root().session().log_warning(
                f"No template named {template_name} in project dependency template"
            )
            return []

        template = templates[template_name]
        defaults = template.get_default_dependency_names()
        kitsu_bindings = project.kitsu_bindings()

        predictive_files = []

        for asset_name in kitsu_casting:
            asset_oid = kitsu_bindings.get_asset_oid(asset_name)
            asset_files = template.get_asset_files(asset_name)

            for name, data in asset_files.items():
                file_type = File if '.' in name else Folder
                pred_file = file_type(asset_name, asset_oid, data['department'], name, data['revision'], project)
                predictive_files.append(pred_file)

        for default in defaults:
            files = template.get_dependency_files(default)

            for name, data in files.items():
                file_type = File if '.' in name else Folder
                pred_file = file_type(default, entity_oid, data['department'], name, data['revision'], project)
                predictive_files.append(pred_file)

    return resolve_files(predictive_files, real_files, site_name)


class DependencyItem(flow.Object):

    ICON = ('icons.flow', 'expanded')

    _dependency_view = flow.Parent()
    dependency_name = flow.Computed(store_value=False)
    entity_name = flow.Computed(store_value=False)
    file_name = flow.Computed(store_value=False)
    revision = flow.Computed(store_value=False)
    revision_oid = flow.Computed(store_value=False)
    status = flow.Computed(store_value=False)
    source_site = flow.Computed(store_value=False)
    comment = flow.Computed(store_value=False)
    path = flow.Computed(store_value=False)
    in_breakdown = flow.Computed(store_value=False)
    is_alien = flow.Computed(store_value=False)

    def compute_child_value(self, child_value):
        if child_value is self.dependency_name:
            self.dependency_name.set(self.name())
        else:
            child_value.set('')

    def color(self):
        return '#444'


class DependencyItemFile(DependencyItem):

    ICON = ('icons.libreflow', 'blank')

    _parent = flow.Parent(3)

    def _get_comment(self, data):
        is_real = data['is_real']
        in_breakdown = data['in_breakdown']
        source_site = data['source_site']
        last_revision = data['last_revision']
        status = self.status.get()
        comment = ""
        
        parent_name = self._parent.name()
        if self.root().session().cmds.Flow.exists(self._parent.oid() + '/display_name'):
            parent_name = self._parent.display_name.get()

        if not self.is_alien.get():
            if not in_breakdown:
                comment = "Not in breakdown"
            elif not is_real:
                comment = "In breakdown but not used in " + parent_name
            elif status == 'Requested' or status == 'Requestable':
                comment = "From " + source_site
            elif self.revision.get() != last_revision:
                comment = "Update available (%s)" % last_revision
        else:
            comment = "In " + parent_name + " but out of pipeline"

        return comment

    def request(self, requesting_site_name=None):
        revision_oid = self.revision_oid.get()
        revision = self.root().get_object(revision_oid)

        if requesting_site_name is None:
            requesting_site_name = self.root().project().get_current_site().name()
        
        request_action = revision.request_as
        request_action.sites.target_site.set(requesting_site_name)
        request_action.sites.source_site.set(self.source_site.get())
        request_action.run(None)

    def compute_child_value(self, child_value):
        data = self._dependency_view.get_dependency_data(self.name())

        if child_value is self.dependency_name:
            self.dependency_name.set(self.file_name.get())
        elif child_value is self.entity_name:
            self.entity_name.set(data['entity_name'])
        elif child_value is self.file_name:
            if self.is_alien.get():
                self.file_name.set(data['path'])
            else:
                self.file_name.set(data['file_name'])
        elif child_value is self.revision:
            self.revision.set(data['revision'] or '-')
        elif child_value is self.revision_oid:
            self.revision_oid.set(data['revision_oid'])
        elif child_value is self.status:
            self.status.set(data['status'])
        elif child_value is self.source_site:
            self.source_site.set(data['source_site'])
        elif child_value is self.comment:
            self.comment.set(self._get_comment(data))
        elif child_value is self.path:
            self.path.set(data['path'])
        elif child_value is self.in_breakdown:
            self.in_breakdown.set(data['in_breakdown'])
        elif child_value is self.is_alien:
            self.is_alien.set(data['is_alien'])

    def color(self):
        return 'default'


class DependencyView(flow.DynamicMap):

    _action = flow.Parent()
    _parent = flow.Parent(2)
    _refresh_time = flow.SessionParam(60).ui(editor='int')
    
    _STATUS_ICONS = {
        'Available': ('icons.libreflow', 'available'),
        'Downloadable': ('icons.libreflow', 'downloadable'),
        'Requestable': ('icons.libreflow', 'requestable'),
        'Requested': ('icons.libreflow', 'waiting'),
        'NotAvailable': ('icons.libreflow', 'unavailable'),
    }

    def __init__(self, parent, name):
        super(DependencyView, self).__init__(parent, name)
        self._deps_time = time.time()
        self._deps_data = None

    @classmethod
    def mapped_type(cls):
        return DependencyItem
    
    def get_site_name(self):
        return self._action.requesting_site.get()
    
    def get_revision_name(self):
        return self._action.revision.get()

    def mapped_names(self, page_num=0, page_size=None):
        self._ensure_dependencies_data()
        mapped_names = []
        
        for entity_name, deps_names in self._entity_deps.items():
            mapped_names.append(entity_name)
            
            for dep_name in deps_names:
                mapped_names.append(dep_name)

        return mapped_names

    def _get_mapped_item_type(self, mapped_name):
        if re.match(r'.*_\d\d\d$', mapped_name):
            return DependencyItemFile
        else:
            return self.mapped_type()

    def columns(self):
        return ['Dependency', 'Revision', 'Comment']

    def get_dependency_data(self, name):
        self._ensure_dependencies_data()
        return self._deps_data[name]

    def update_dependencies_data(self):
        include_real = not self._action.predictive_only.get()
        deps = sorted(
            get_dependencies(
                self._parent,
                site_name=self.get_site_name(),
                revision_name=self.get_revision_name(),
                predictive=True,
                real=include_real,
            ),
            key=lambda d: d['is_alien']
        )
        
        files_data = defaultdict(list)
        self._deps_data = dict()
        self._entity_deps = defaultdict(list)
        
        for d in deps:
            files_data[d['entity_name']].append(d)
        
        for entity_name, data in files_data.items():
            for i, file_data in enumerate(data):
                dep_name = entity_name + '_%03i' % i
                self._entity_deps[entity_name].append(dep_name)
                self._deps_data[dep_name] = file_data

    def _ensure_dependencies_data(self):
        if not self._deps_data or time.time() - self._deps_time > self._refresh_time.get():
            self.update_dependencies_data()
            self._deps_time = time.time()

    def _fill_row_cells(self, row, item):
        row['Dependency'] = item.dependency_name.get()
        row['Revision'] = item.revision.get()
        row['Comment'] = item.comment.get()

    def _fill_row_style(self, style, item, row):
        color = item.color()
        if color != 'default':
            for col in self.columns():
                style['%s_background-color' % col] = color
        
        style['icon'] = self._STATUS_ICONS.get(item.status.get(), ('icons.flow', 'expanded'))


class WorkingSiteName(flow.values.ChoiceValue):
    
    STRICT_CHOICES = False
    
    def choices(self):
        return self.root().project().get_working_sites().mapped_names()
    
    def revert_to_default(self):
        self.set(self.root().project().get_current_site().name())


class RevisionName(flow.values.ChoiceValue):
    
    _parent = flow.Parent(2)
    STRICT_CHOICES = False
    
    def choices(self):
        try:
            revisions = self._parent.get_revision_names(published_only=True)
        except AttributeError:
            revisions = []
        
        return revisions
    
    def revert_to_default(self):
        try:
            last_revision = self._parent.get_head_revision().name()
        except AttributeError:
            last_revision = None
        
        self.set(last_revision)


class GetDependenciesAction(flow.Action):

    ICON = ('icons.libreflow', 'dependencies')

    requesting_site = flow.Param(None, WorkingSiteName).watched()
    revision = flow.Param(None, RevisionName).watched()
    dependencies = flow.Child(DependencyView).ui(expanded=True)
    predictive_only = flow.SessionParam(False).watched().ui(editor='bool')

    def child_value_changed(self, child_value):
        if child_value is self.predictive_only:
            self.dependencies.update_dependencies_data()
            self.dependencies.touch()
        elif child_value in [self.requesting_site, self.revision]:
            self.update_dependencies()
            
    def update_dependencies(self):
        self.dependencies.update_dependencies_data()
        self.dependencies.touch()

    def get_buttons(self):
        self.message.set("<h2>Get dependencies</h2>")
        self.requesting_site.revert_to_default()
        self.revision.revert_to_default()
        
        return ['Proceed', 'Cancel']

    def run(self, button):
        if button == 'Cancel':
            return
        
        for d in self.dependencies.mapped_items():
            if type(d) is DependencyItemFile:
                if d.status.get() in ['Requestable', 'Downloadable']:
                    d.request(self.requesting_site.get())

        self.update_dependencies()
        
        return self.get_result(close=False)


class DependencyStatus(flow.values.ChoiceValue):

    CHOICES = ['Available', 'Downloadable', 'Requestable', 'Unavailable']
    _COLORS = {
        'Available': '#45cc3d',
        'Downloadable': '#ffa800',
        'Requestable': '#aaa',
        'Unavailable': '#cc3b3c'
    }

    @classmethod
    def get_color(cls, status):
        return cls._COLORS[status]


class DependencyFile(flow.Object):

    _dependency = flow.Parent(3)
    department = flow.Computed(store_value=False)
    file_format = flow.Computed(store_value=False)
    revision = flow.Computed(store_value=False)
    status = flow.Computed(store_value=False)

    def get_name(self):
        dependency_files = self._dependency.get_dependency_files()
        name = '.'.join(self.name().rsplit('_', 1))

        if not name in dependency_files:
            name = self.name()

        return name

    def _get_root_oid(self):
        return self._dependency.dependency_oid.get() + '/departments/' + self.department.get() + '/files'

    def get_oid(self):
        return self._get_root_oid() + '/' + self.name()

    def get_file_data(self):
        dependency_files = self._dependency.get_dependency_files()
        return dependency_files[self.get_name()]

    def exists(self):
        return self.root().session().cmds.Flow.exists(self.get_oid())

    def _get_status(self):
        status = 'Unavailable'

        if self.exists():
            revision = self._get_revision()

            if revision:
                if revision.get_sync_status() == 'Available':
                    status = 'Available'
                elif revision.get_sync_status(exchange=True) == 'Available':
                    status = 'Downloadable'
                else:
                    status = 'Requestable'

        return status


    def _get_revision(self):
        revision = None

        if not self.exists():
            return revision

        revision_name = self.revision.get()
        tracked_file = self.root().get_object(self.get_oid())

        if revision_name == 'last':
            revision = tracked_file.get_head_revision()
        elif tracked_file.has_revision(revision_name):
            revision = tracked_file.get_revision(revision_name)

        return revision

    def compute_child_value(self, child_value):
        file_data = self.get_file_data()

        if child_value is self.department:
            self.department.set(file_data['department'])
        elif child_value is self.file_format:
            default_format = self.get_name().split('.')[-1]
            self.file_format.set(file_data.get('format', default_format))
        elif child_value is self.revision:
            self.revision.set(file_data['revision'])
        elif child_value is self.status:
            self.status.set(self._get_status())


class DependencyFileMap(flow.DynamicMap):

    _dependency = flow.Parent(2)

    @classmethod
    def mapped_type(cls):
        return DependencyFile

    def mapped_names(self, page_num=0, page_size=None):
        dependency_files = self._dependency.get_dependency_files()
        names = [name.replace('.', '_') for name in dependency_files]
        return names

    def _fill_row_cells(self, row, item):
        row['Name'] = item.get_name()

    def _fill_row_style(self, style, item, row):
        color = DependencyStatus.get_color(item.status.get())
        style['foreground-color'] = color


class ShowDependencyFiles(flow.Action):

    _dependency = flow.Parent()

    def needs_dialog(self):
        return False

    def run(self, button):
        return self.get_result(
            goto=self._dependency.files.oid()
        )


class DependencyFiles(flow.Object):

    file_map = flow.Child(DependencyFileMap).ui(
        label='Files',
        expanded=True
    )


class Dependency(flow.Object):

    _parent = flow.Parent()

    files = flow.Child(DependencyFiles)
    dependency_oid = flow.Computed(store_value=False)
    nb_occurrences = flow.Computed(store_value=False).ui(editor='int')
    available = flow.Computed(store_value=False).ui(editor='bool')

    show_files = flow.Child(ShowDependencyFiles)

    def get_dependency_data(self):
        return self._parent.get_asset_data(self.name())

    def get_dependency_files(self):
        dependency_data = self.get_dependency_data()
        return dependency_data['files']

    def get_files(self):
        return self.files.file_map.mapped_items()

    def get_file_count(self, status):
        file_statuses = [f.status.get() for f in self.get_files()]
        return file_statuses.count(status)

    def compute_child_value(self, child_value):
        dependency_data = self.get_dependency_data()

        if child_value is self.dependency_oid:
            self.dependency_oid.set(dependency_data['dependency_oid'])
        elif child_value is self.nb_occurrences:
            self.nb_occurrences.set(dependency_data['nb_occurrences'])
        elif child_value is self.available:
            available = (self.get_file_count('Available') == len(self.get_files()))
            self.available.set(available)


class Dependencies(flow.DynamicMap):

    _shot = flow.Parent()
    _sequence = flow.Parent(3)
    _updated = flow.BoolParam(False)
    _refresh_time = flow.SessionParam(30).ui(editor='int')

    def __init__(self, parent, name):
        super(Dependencies, self).__init__(parent, name)
        self._assets_data_time = time.time()
        self._assets_data = None

    def mapped_names(self, page_num=0, page_size=None):
        self._ensure_assets_data()
        return list(self._assets_data.keys())

    def _ensure_assets_data(self):
        if not self._assets_data or time.time() - self._assets_data_time > self._refresh_time.get():
            self._assets_data = self._get_dependencies_data()
            self._assets_data_time = time.time()

    def _get_dependencies_data(self):
        kitsu_api = self.root().project().kitsu_api()
        kitsu_casting = kitsu_api.get_shot_casting(self._shot.name(), self._sequence.name())
        kitsu_bindings = self.root().project().admin.kitsu.bindings
        assets_data = dict()

        dependency_template = self.root().project().admin.dependency_templates['shot']
        dependencies = dependency_template.get_dependencies()
        defaults = dependency_template.get_default_dependency_names()

        # Kitsu dependencies
        for asset in kitsu_casting:
            asset_name = asset['asset_name']
            asset_type = asset['asset_type_name']
            asset_oid = kitsu_bindings.get_asset_oid(asset_name)

            files = dependencies['kitsu.'+asset_type]

            assets_data[asset_name] = dict(
                dependency_oid=asset_oid,
                files=files,
                nb_occurrences=asset['nb_occurences'],
            )

        # Default dependencies
        for default in defaults:
            files = dependencies[default]

            assets_data[default] = dict(
                dependency_oid=self._shot.oid(),
                files=files,
                nb_occurrences=1,
            )

        return assets_data

    def get_asset_data(self, asset_name):
        self._ensure_assets_data()
        return self._assets_data[asset_name]

    @classmethod
    def mapped_type(cls):
        return Dependency

    def columns(self):
        return ['Name', 'Occurrences'] + DependencyStatus.CHOICES

    def _fill_row_cells(self, row, item):
        row['Name'] = item.name()
        row['Occurrences'] = item.nb_occurrences.get()

        for status in DependencyStatus.CHOICES:
            file_count = item.get_file_count(status)
            if file_count:
                row[status] = str(file_count)
            else:
                row[status] = ""

    def _fill_row_style(self, style, item, row):
        if item.available.get():
            style['foreground-color'] = '#45cc3d'
        else:
            style['foreground-color'] = '#cc3b3c'

        for status in DependencyStatus.CHOICES:
            color = DependencyStatus.get_color(status)
            style['%s_foreground-color' % status] = color



import os
import shutil

from kabaret import flow
from kabaret.flow_contextual_dict import ContextualView, get_contextual_dict

from .departments import Department
from .maputils import (
    GenericItemMap,
    CreateItemAction,
    CreateGenericItemAction,
    ClearMapAction,
)
from .kitsu import KitsuAsset, KitsuMap, UpdateItemsKitsuSettings


class CreateAssetAction(CreateGenericItemAction):
    """
    Adds an asset to a map given its name.
    """

    _map = flow.Parent()

    asset_name = flow.Param("").ui(label="Name")

    def needs_dialog(self):
        return True

    def run(self, button):
        if button == "Cancel":
            return

        item = self._map.add(self.asset_name.get())

        if item is None:
            msg = self.message.get()
            msg += "\n<h3>Could not create asset <i>{}</i>...</h3>".format(
                self.asset_name.get()
            )
            msg += "\nSee logs for details"
            self.message.set(msg)

            return self.get_result(close=False)

        self._map.touch()


class AssetDepartments(flow.Object):

    model = flow.Child(Department).ui(expanded=True)
    rig = flow.Child(Department).ui(expanded=True)


class Asset(KitsuAsset):
    """
    Abstraction of an asset.
    """

    ICON = ("icons.flow", "asset")

    _assets = flow.Parent()

    settings = flow.Child(ContextualView).ui(hidden=True)
    description = flow.Param("")

    departments = flow.Child(AssetDepartments).ui(expanded=True)

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(asset_name=self.name())


class Assets(GenericItemMap):
    """
    Map of assets.
    """

    ICON = ("icons.flow", "bank")

    item_prefix = "Asset"

    create_asset = flow.Child(CreateAssetAction)

    with flow.group("Kitsu"):
        update_kitsu_settings = flow.Child(UpdateItemsKitsuSettings)

    @classmethod
    def mapped_type(cls):
        return Asset

    def _fill_row_cells(self, row, item):
        row["Name"] = item.name()
        row["Description"] = item.description.get()

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            # return dict(path="LIB", file_category="LIB")
            return dict(file_category="LIB")


class AssetFamily(flow.Object):

    _asset_type = flow.Parent(2)

    description = flow.Param("")
    assets = flow.Child(Assets)

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(asset_family=self.name().lower())


class AssetFamilies(flow.Map):

    create_asset_family = flow.Child(CreateItemAction)

    @classmethod
    def mapped_type(cls):
        return AssetFamily


class AssetType(flow.Object):

    _lib = flow.Parent()

    description = flow.Param("")
    asset_families = flow.Child(AssetFamilies)

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(asset_tpye=self.name().lower())


class AssetTypes(flow.Map):

    create_asset_type = flow.Child(CreateItemAction)

    @classmethod
    def mapped_type(cls):
        return AssetType


class AssetLib(flow.Object):

    asset_types = flow.Child(AssetTypes)

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(file_category="LIB")


# Dependencies
# ----------------------


class DropAssetAction(flow.ConnectAction):
    """
    Adds an asset dependency in a map by drag-and-dropping an Asset.
    """

    _map = flow.Parent()

    def accept_label(self, objects, urls):
        asset_objects = list(filter(lambda obj: isinstance(obj, Asset), objects))
        return super(DropAssetAction, self).accept_label(asset_objects, urls)

    def _add_asset(self, asset):
        asset_ref = self._map.add(asset.name())
        asset_ref.set(asset)

    def run(self, objects, urls):
        for obj in objects:
            self._add_asset(obj)

        self._map.touch()


class AssetDependency(flow.values.Ref):
    """
    Abstraction of an asset dependency.
    """

    ICON = ("icons.flow", "casting")

    SOURCE_TYPE = Asset


class AssetDependencies(flow.Map):
    """
    Map of asset dependencies.
    """

    ICON = ("icons.flow", "casting")

    drop_asset = flow.Child(DropAssetAction)
    clear_map = flow.Child(ClearMapAction)

    @classmethod
    def mapped_type(cls):
        return AssetDependency

    def columns(self):
        return ["Name", "Description"]

    def row(self, item):
        _, row = super(AssetDependencies, self).row(item)
        asset = item.get()

        return asset.oid(), row

    def _fill_row_cells(self, row, item):
        asset = item.get()
        row["Name"] = asset.id.get()
        row["Description"] = asset.description.get()



from kabaret import flow

from libreflow import baseflow, examples
from .film import Episodes, Shots
from .lib import AssetLib
from .file import CreateTrackedFileAction, FileSystemMap, Revisions
from .kitsu import SyncFromKitsu
from .runners import (
    AfterEffect,
    AfterEffectRender,
    Photoshop,
)
from .users import Bookmark


class Project(baseflow.Project, flow.InjectionProvider):
    episodes = flow.Child(Episodes).ui(default_height=420, expanded=True)
    asset_lib = flow.Child(AssetLib).ui(expanded=True)
    admin = flow.Child(baseflow.Admin)

    sequences = flow.Child(baseflow.film.Sequences).ui(hidden=True)

    def _register_runners(self):
        super(Project, self)._register_runners()
        self._RUNNERS_FACTORY.ensure_runner_type(AfterEffect)
        self._RUNNERS_FACTORY.ensure_runner_type(AfterEffectRender)
        self._RUNNERS_FACTORY.ensure_runner_type(Photoshop)

    @classmethod
    def _injection_provider(cls, slot_name, default_type):
        if slot_name == "libreflow.baseflow.file.CreateTrackedFileAction":
            return CreateTrackedFileAction
        elif slot_name == "libreflow.baseflow.kitsu.SyncFromKitsu":
            return SyncFromKitsu
        elif slot_name == "libreflow.baseflow.file.FileSystemMap":
            return FileSystemMap
        elif slot_name == "libreflow.baseflow.file.Revisions":
            return Revisions
        elif slot_name == "libreflow.baseflow.users.Bookmark":
            return Bookmark

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            defaultContextualEdits = baseflow.Project.get_default_contextual_edits(
                self, "settings"
            )
            defaultContextualEdits.update(
                dict(
                    default_shot_layout_files="@{sequence}_{shot}_layout.psd, @{sequence}_{shot}_layout.aep",
                    default_shot_animation_files="{sequence}_{shot}_animationB.blend, {sequence}_{shot}_animationA.aep, @render, @preview",
                    default_shot_compositing_files="{sequence}_{shot}_compositing.aep, @render",
                    default_asset_design_files="{asset_type}_{asset_name}_design.psd,{asset_type}_{asset_name}_design.png, @ref",
                    default_asset_cycle_files="{asset_type}_{asset_name}_cycle.blend, @render, @preview",
                )
            )
            return defaultContextualEdits



#!/usr/bin/env python3

# Mark sequence, copyright (C) 2020 Les Fées Spéciales
# voeu@les-fees-speciales.coop
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import os
import platform
import subprocess
import json
import argparse
import fileseq
import textwrap
from tempfile import mkdtemp
from math import inf
from concurrent.futures import ThreadPoolExecutor


__all__ = ['default_template', 'SequenceMarker']

# Paths of installed binaries, depending on environment
# It would be better to let the OS handle the path, but for now let's not
bins = {'Linux':
        {'ffmpeg': 'ffmpeg',
         'convert': 'convert'},
        'Windows':
        {'ffmpeg': 'ffmpeg',
         'convert': 'convert'},
        'Darwin':
        {'ffmpeg': 'ffmpeg',
         'convert': 'convert'}
}


default_template = {
    "settings": {
        "font_size": 24,
        "color": "white"
        # "color": "chartreuse"
    },
    "fields": [
        {
            "name": "project",
            "direction": "NorthWest",
            "string": " %s "
        },
        {
            "name": "sequence",
            "direction": "NorthWest",
            "string": "%s "
        },
        {
            "name": "scene",
            "direction": "NorthWest",
            "string": "%s "
        },
        {
            "name": "frame_number",
            "direction": "NorthWest",
            "string": "%s"
        },
        {
            "name": "normalized_frame_number",
            "direction": "North",
            "string": "%04i / "
        },
        {
            "name": "total_images",
            "direction": "North",
            "string": "%s"
        },
        {
            "name": "file_name",
            "direction": "NorthEast",
            "string": " %s "
        },
        {
            "name": "version",
            "direction": "NorthEast",
            "string": " %s "
        },
        {
            "name": "resolution",
            "direction": "NorthEast",
            "string": " %s "
        },
        {
            "name": "copyright",
            "direction": "SouthWest",
            "string": " %s "
        },
        {
            "name": "simplify",
            "direction": "SouthWest",
            "string": " %s "
        },
        {
            "name": "focal_length",
            "direction": "SouthWest",
            "string": " Focal length: %d mm "
        },
        {
            "name": "studio",
            "direction": "SouthEast",
            "string": " %s "
        },
        {
            "name": "user",
            "direction": "SouthEast",
            "string": " %s "
        },
        {
            "name": "hostname",
            "direction": "SouthEast",
            "string": " %s "
        },
        {
            "name": "date",
            "direction": "SouthEast",
            "string": " %s "
        }
    ],
    "image_fields": [
        # {
        #     "name": "circle",
        #     "direction": "SouthWest",
        #     "geometry": "10x10+20+4"
        # }
    ]
}

class SequenceMarker():
    def __init__(self, image_filepath, data, template=default_template):
        self.data = data
        self.template = template
        self.create_temp_dir()

        self.file_sequence = fileseq.findSequenceOnDisk(image_filepath)
        self.frame_set = self.file_sequence.frameSet()

    def create_temp_dir(self):
        """Create temporary directory for images"""
        if 'mark_dir' in self.data and self.data['mark_dir']:
            self.mark_dir = self.data['mark_dir']
            os.makedirs(self.mark_dir, exist_ok=True)
        else:
            self.mark_dir = mkdtemp()

    def delete_temp_dir(self):
        """Delete temporary directory"""
        if not ('mark_dir' in self.data and self.data['mark_dir']):
            print("Deleting temp dir...")
            from shutil import rmtree
            rmtree(self.mark_dir)

    def mark_sequence(self):
        last_image_marked = self.mark_images()

        marked_sequence = fileseq.findSequenceOnDisk(last_image_marked)
        if self.data['video_output']:
            self.render_video(self.get_sequence_path(marked_sequence),
                              os.path.abspath(self.data['video_output']),
                              audio_file=self.data["audio_file"],
                              frame_rate=self.data["frame_rate"])

        self.delete_temp_dir()

    def mark_images(self):
        """Batch mark images"""
        image_data = self.data.copy()

        # Special fields: for each special field, give a default if it is
        # specified in the template but not passed as data
        for field in self.template['fields']:
            if field['name'] == 'date':
                import datetime
                image_data['date'] = datetime.datetime.now().strftime("%d-%m-%y %H:%M")
            if field['name'] == 'user':
                import getpass
                image_data['user'] = getpass.getuser()
            if field['name'] == 'hostname':
                import platform
                image_data['hostname'] = platform.node()
            if field['name'] == 'total_images':
                image_data['total_images'] = len(self.frame_set)

        with ThreadPoolExecutor() as executor:
            for i, image_number in enumerate(self.frame_set):
                if (image_number < self.data['start_frame']
                    or image_number > self.data['end_frame']):
                    continue
                image_source = self.file_sequence.frame(image_number)
                image_marked = os.path.join(self.mark_dir,
                                            "marked.%04i.tif" % (i - self.data['offset'] + 1))

                # Special fields evaluated at each frame
                for field in self.template['fields']:
                    if field['name'] == 'frame_number':
                        image_data['frame_number'] = image_number
                    if field['name'] == 'normalized_frame_number':
                        image_data['normalized_frame_number'] = i - self.data['offset'] + 1
                    if field['name'] in self.data and type(self.data[field['name']]) is dict:
                        image_data[field['name']] = self.data[field['name']][image_number]

                executor.submit(self.mark_image, image_source, image_marked, image_data.copy())

        # Return last image path
        return image_marked

    def mark_image(self, path, output_path, image_data):
        '''Use ImageMagick's convert command line utility to overlay metadata on
        specified image'''
        print("Marking image %s..." % image_data['frame_number'])

        convert_bin = bins[platform.system()]['convert']
        convert_args = [convert_bin]
        convert_args += ['%s' % path]

        settings = self.template['settings']

        # Add gray band overlay
        convert_args.extend(['-fill', 'rgba(0,0,0,0.5)'])
        convert_args.extend(['-draw', 'rectangle 0,0 %s,%s' % (image_data['resolution_x'],
                                                               settings['font_size'])])
        convert_args.extend(['-draw', 'rectangle 0,%s %s,%s' % (image_data['resolution_y'] -
                                                                settings['font_size'] - 2,
                                                                image_data['resolution_x'],
                                                                image_data['resolution_y'])])

        directions = {}

        # Add annotations for each field to the list of directions
        # This has the effect of concatenating various fields for a given direction
        for field in self.template['fields']:
            direction = field['direction']
            value = field['string']
            # Try formatting the string with the value from the passed data
            try:
                value %= image_data[field['name']]
            except BaseException as e:
                print(f"Could not evaluate field {field['name']}: {e}")
                continue
            if not direction in directions:
                directions[direction] = ''
            directions[direction] += (value)

        # Set text color and size for outside stroke
        convert_args.extend(['-fill', 'black', '-strokewidth', '3',
                             '-stroke', 'black', '-weight', 'bold',
                             '-pointsize', str(settings['font_size'])])

        # Add annotations for each field, for outside stroke
        for direction, value in directions.items():
            convert_args.extend(['-gravity', direction, '-annotate', '0', value])

        # Set text color and size for fill
        convert_args.extend(['-fill', settings['color'],
                             '-stroke', 'none', '-weight', 'bold',
                             '-pointsize', str(settings['font_size'])])

        # Add annotations for each field, with only inside fill
        for direction, value in directions.items():
            convert_args.extend(['-gravity', direction, '-annotate', '0', value])

        # Add image annotations
        for image in self.template['image_fields']:
            convert_args.append('(')

            # File path, either from template or from command line
            if image['field'] and image_data[image['field']]:
                convert_args.append(os.path.abspath(image_data[image['field']]))
            else:
                convert_args.append(image['path'])
            convert_args.extend([
                '-gravity', image['direction'],
                '-geometry', image['geometry'],
                ')',
                '-composite'])

        # Debug alpha channel
        convert_args.extend(['-alpha', 'remove'])
        convert_args.extend(['-compress', 'Piz'])  # TODO : remettre DWAA quand ffmpeg le permettra

        # Output
        convert_args.append('%s' % output_path)
        # Windows needs to use shell=True, see https://stackoverflow.com/a/41860823
        proc = subprocess.run(convert_args, check=True, shell=platform.system() == 'Windows')

    def render_video(self, img_sources, destination, audio_file=None, frame_rate=25):
        print("Generating video...")
        ffmpeg_bin = bins[platform.system()]['ffmpeg']
        ffmpeg_args = [ffmpeg_bin, '-y', '-loglevel', 'error']
        ffmpeg_args.extend(['-r', str(frame_rate)])
        ffmpeg_args.extend(['-i', img_sources])

        if audio_file is not None:
            ffmpeg_args.extend(['-i', audio_file])
            ffmpeg_args.extend(['-c:a', 'aac', "-b:a", "160k"])
            ffmpeg_args.extend(["-map", "0:0", "-map", "1:0"])

        ffmpeg_args.extend(['-c:v', 'mjpeg', '-q:v', '3'])
        # ffmpeg_args.extend(['-c:v', 'h264', '-crf', '25', '-preset', 'slow', '-pix_fmt', 'yuv420p'])

        os.makedirs(os.path.dirname(destination), exist_ok=True)
        ffmpeg_args.extend(['%s' % (destination)])

        proc = subprocess.run(ffmpeg_args)

    @staticmethod
    def get_sequence_path(sequence):
        padding = sequence.getPaddingNum(sequence.padding())
        return sequence.format('{dirname}{basename}%0' + str(padding) + 'd{extension}')


if __name__ == "__main__":
    # Parse command line arguments
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.indent(
            textwrap.dedent('''\

            Make an annotated movie output from a list of images. A JSON
            template may be specified, which will contain fields such as:

            {
                "name": "scene",
                "direction": "NorthWest",
                "string": ' sc%s '
            },

            You can then specify the option --scene on the command line, and
            the text will appear in the top left. Warning: underscores are
            replaced by dashes, so "my_field" becomes "my-field", to respect
            the customary option format.

            The direction uses ImageMagick’s convention: Center, North,
            NorthEast, East, SouthEast, South, SouthWest, West, NorthWest. If a
            direction is specified multiple times, the corresponding fields
            will be concatenated.

            '''
            ), '  '
        )
    )

    group = parser.add_argument_group('file options')
    group.add_argument('-t', '--template', type=str,
                        help='template file containing field descriptions')

    group.add_argument('image_sequence', type=str,
                        help='input image sequence, typically a frame in the sequence')
    group.add_argument('-d', '--mark-dir', type=str,
                        help='intermediate directory, leave blank for tmp dir')
    group.add_argument('-o', '--video-output', type=str,
                        help='render video to this destination')
    group.add_argument('-a', '--audio-file', type=str,
                        help='if rendering video, use this file as audio track')

    group.add_argument('-r', '--frame_rate', type=float, default=25.0,
                        help='playback speed')

    group = parser.add_argument_group('frame options')
    group.add_argument('-O', '--offset', type=int, default=0,
                        help='offset for renaming frames')
    group.add_argument('-s', '--start-frame', type=int, default=-inf,
                        help="don't mark images lower than this number")
    group.add_argument('-e', '--end-frame', type=int, default=inf,
                        help="don't mark images higher than this number")

    args = parser.parse_known_args()[0]

    # Load in template from supplied json file. If none given, use default one.
    if args.template is None:
        template = default_template
    else:
        with open(os.path.abspath(args.template), 'r') as f:
            template = json.load(f)

    # Add text fields to argument parser
    group = parser.add_argument_group('Template text field arguments')
    for field in template['fields']:
        field = field['name'].replace('_', '-')
        group.add_argument('--' + field, type=str, default='')

    # Add image fields to argument parser
    group = parser.add_argument_group('Template image field arguments')
    for image in template['image_fields']:
        image = image['name'].replace('_', '-')
        group.add_argument('--' + image, type=str, default='')

    args = parser.parse_args()

    sequence_marker = SequenceMarker(os.path.abspath(args.image_sequence), vars(args), template=template)

    # Get resolution from first image
    res = subprocess.check_output(['identify', '-format', '%wx%h',
                                   sequence_marker.file_sequence.frame(sequence_marker.frame_set[0])])
    res_x, res_y = res.decode('ascii').split("x")
    sequence_marker.data['resolution_x'] = int(res_x)
    sequence_marker.data['resolution_y'] = int(res_y)
    
    # If not specified, update sequence resolution with that of the first frame
    if not 'resolution' in sequence_marker.data:
        sequence_marker.data['resolution'] = res.decode('ascii')

    sequence_marker.mark_sequence()
